/* Cairo Gtk Engine for GTK+ 2.x */

#include "cge.h"


#define DETAIL(xx) ((detail) && (!strcmp(xx, detail)))


static void color_gdk_to_rgb(GdkColor  color,gdouble *r,gdouble *g,gdouble *b);
static GdkColor color_tint_black(GdkColor color,gdouble t); 
/* static GdkColor color_tint_white(GdkColor color,gdouble t); */ 
static void gdk_get_off(GdkDrawable *window,gint* x_off,gint* y_off); 
static void cge_gdk(cairo_t *xrs, GdkDrawable *window); 
static void cge_draw_line(cairo_t *xrs, GdkColor c, gdouble alpha, 
			     double x1, double y1, 
			     double x2, double y2, 
			     gboolean line_width); 
static cge_style *cge_get_style(GtkStyle *style,const gchar *detail);

static cairo_surface_t *cge_create_gradient(cairo_t *xrs,
					    gint ncols,
					    GdkColor *colors,
					    gdouble alpha,
					    cge_direction d,
					    gint x_off,
					    gint y_off,
					    gint width,
					    gint height);

/*
static void cge_draw_slider(cairo_t *xrs,
                               GdkColor color,
                               gdouble alpha,
                               gint x,gint y,gint w, gint h,
                               gdouble line_width,
                               gint line_count
                               );
*/

static void cge_draw_checked(cairo_t *xrs,
			     GdkColor color,
			     gdouble alpha,
			     gint x,gint y,gint w, gint h,
			     gdouble line_width
			     );

static void cge_draw_arrow_from_rc(cairo_t *xrs,
			   GtkStyle * style,
			   GtkStateType state_type,
			   const gchar *detail,
			   GdkColor c,
			   gdouble alpha,
			   gint x,gint y,gint w,gint h,
			   gboolean fill,
			   gdouble line_width,
			   GtkArrowType at);

static void cge_draw_rectangle(cairo_t *xrs,
			       GdkColor c,
			       gdouble alpha,
			       gint x,
			       gint y,
			       gint w,
			       gint h);

static void cge_draw_box(cairo_t *xrs,
                            GdkColor c,
                            cairo_surface_t *pattern,
                            gdouble alpha,
                            gdouble x,
                            gdouble y,
                            gdouble w,
                            gdouble h,
                            gdouble round_size,
                            gboolean fill,
                            gdouble line_width);

static void cge_draw_circle(cairo_t *xrs,
                               GdkColor c,
                               cairo_surface_t *pattern,
                               gdouble alpha,
                               gint x,
                               gint y,
                               gdouble round_size,
                               gboolean fill,
                               gdouble line_width);

static void cge_draw_box_gap(cairo_t *xrs,
                                GdkColor c,
                                cairo_surface_t *pattern,
                                gdouble alpha,
                                gdouble x,
                                gdouble y,
                                gdouble w,
                                gdouble h,
                                gdouble round_size,
                                gint gap_x,
                                gint gap_width,
                                gboolean fill,
                               gdouble line_width);

static void cge_draw_notebook_tab(cairo_t *xrs,
                                     GdkColor c,
                                     cairo_surface_t *pattern,
                                     gdouble alpha,
                                     gdouble x,
                                     gdouble y,
                                     gdouble w,
                                     gdouble h,
                                     gdouble round_size,
                                     gboolean fill,
                                     gdouble line_width);

static cairo_surface_t *cge_create_gradient_from_rc(cairo_t *xrs,
                                                       GtkStyle * style,
                                                       GtkStateType state_type,
                                                       const gchar *detail,
                                                       gint x_off,
                                                       gint y_off,
                                                       gint width,
                                                       gint height);

static void cge_draw_filled_box_from_rc(cairo_t *xrs,
                                           GtkStyle * style,
                                           GtkStateType state_type,
                                           const gchar *detail,
                                           gint x_off,
                                           gint y_off,
                                           gint x,
                                           gint y,
                                           gint width,
                                           gint height);

static void cge_draw_unfilled_box_from_rc(cairo_t *xrs,
                                             GtkStyle * style,
                                             GtkStateType state_type,
                                             const gchar *detail,
                                             gint x,
                                             gint y,
                                             gint width,
                                             gint height);

static void cge_draw_line_from_rc(cairo_t *xrs,
                                     GtkStyle * style,
                                     GtkStateType state_type,
                                     const gchar *detail,
                                     double x1,
                                     double y1,
                                     double x2,
                                     double y2); 



static void cge_style_init (CgeStyle * style);
static void cge_style_class_init (CgeStyleClass * klass);


static void draw_hline (GtkStyle * style,
			GdkWindow * window,
			GtkStateType state_type,
			GdkRectangle * area,
			GtkWidget * widget,
			const gchar * detail, gint x1, gint x2, gint y);

static void draw_vline (GtkStyle * style,
			GdkWindow * window,
			GtkStateType state_type,
			GdkRectangle * area,
			GtkWidget * widget,
			const gchar * detail, gint y1, gint y2, gint x);

static void draw_tab (GtkStyle *style,
		      GdkWindow *window,
		      GtkStateType state_type,
		      GtkShadowType shadow_type,
		      GdkRectangle *area,
		      GtkWidget *widget,
		      const gchar *detail,
		      gint x,
		      gint y,
		      gint width,
		      gint height);

static void draw_shadow (GtkStyle * style,
			 GdkWindow * window,
			 GtkStateType state_type,
			 GtkShadowType shadow_type,
			 GdkRectangle * area,
			 GtkWidget * widget,
			 const gchar * detail,
			 gint x, gint y, gint width, gint height);

static void draw_shadow_gap (GtkStyle               *style,
			     GdkWindow              *window,
			     GtkStateType            state_type,
			     GtkShadowType           shadow_type,
			     GdkRectangle           *area,
			     GtkWidget              *widget,
			     const gchar            *detail,
			     gint                    x,
			     gint                    y,
			     gint                    width,
			     gint                    height,
			     GtkPositionType         gap_side,
			     gint                    gap_x,
			     gint                    gap_width);



static void draw_arrow (GtkStyle * style,
			GdkWindow * window,
			GtkStateType state_type,
			GtkShadowType shadow_type,
			GdkRectangle * area,
			GtkWidget * widget,
			const gchar * detail,
			GtkArrowType arrow_type,
			gint fill, gint x, gint y, gint width, gint height);

static void draw_box (GtkStyle * style,
		      GdkWindow * window,
		      GtkStateType state_type,
		      GtkShadowType shadow_type,
		      GdkRectangle * area,
		      GtkWidget * widget,
		      const gchar * detail,
		      gint x, gint y, gint width, gint height);

static void draw_box_gap (GtkStyle *style,
			  GdkWindow *window,
			  GtkStateType state_type,
			  GtkShadowType shadow_type,
			  GdkRectangle *area,
			  GtkWidget *widget,
			  const gchar *detail,
			  gint x,
			  gint y,
			  gint width,
			  gint height,
			  GtkPositionType gap_side,
			  gint gap_x,
			  gint gap_width);

static void draw_flat_box (GtkStyle * style,
			   GdkWindow * window,
			   GtkStateType state_type,
			   GtkShadowType shadow_type,
			   GdkRectangle * area,
			   GtkWidget * widget,
			   const gchar * detail,
			   gint x, gint y, gint width, gint height);

static void draw_check (GtkStyle * style,
			GdkWindow * window,
			GtkStateType state_type,
			GtkShadowType shadow_type,
			GdkRectangle * area,
			GtkWidget * widget,
			const gchar * detail,
			gint x, gint y, gint width, gint height);

static void draw_option (GtkStyle * style,
			 GdkWindow * window,
			 GtkStateType state_type,
			 GtkShadowType shadow_type,
			 GdkRectangle * area,
			 GtkWidget * widget,
			 const gchar * detail,
			 gint x, gint y, gint width, gint height);


static void draw_extension (GtkStyle * style,
			    GdkWindow * window,
			    GtkStateType state_type,
			    GtkShadowType shadow_type,
			    GdkRectangle * area,
			    GtkWidget * widget,
			    const gchar * detail,
			    gint x,
			    gint y,
			    gint width,
			    gint height, GtkPositionType gap_side);

static void draw_slider (GtkStyle * style,
			 GdkWindow * window,
			 GtkStateType state_type,
			 GtkShadowType shadow_type,
			 GdkRectangle * area,
			 GtkWidget * widget,
			 const gchar * detail,
			 gint x,
			 gint y,
			 gint width, gint height, GtkOrientation orientation);

static void draw_handle (GtkStyle * style,
			 GdkWindow * window,
			 GtkStateType state_type,
			 GtkShadowType shadow_type,
			 GdkRectangle * area,
			 GtkWidget * widget,
			 const gchar * detail,
			 gint x,
			 gint y,
			 gint width, gint height, GtkOrientation orientation);

static void draw_resize_grip (GtkStyle * style, GdkWindow * window,
			      GtkStateType state_type, GdkRectangle * area,
			      GtkWidget * widget, const gchar * detail,
			      GdkWindowEdge edge, gint x, gint y,
			      gint width, gint height);


static GtkStyleClass *parent_class = NULL;

static cge_style *cge_get_style(GtkStyle *style,const gchar *detail)
{
  gint i;


  for(i=1;i<MAX_STYLES;i++)       
    {
      cge_style *cstyle=&THEME_DATA(style)->styles[i];
      if(cstyle->detail!=NULL && detail!=NULL)     
      {
	//printf("%s - %s\n",cstyle->detail,detail);
	if(g_ascii_strncasecmp(cstyle->detail, detail, strlen(detail))==0)  
	  {
	    //printf("style %s found\n",detail);
	    return cstyle;
	  }
      }
    }
  return &THEME_DATA(style)->styles[0];
}


static void color_gdk_to_rgb(GdkColor  color,gdouble *r,gdouble *g,gdouble *b)
{
	*r=(double)color.red/(double)65535;
	*g=(double)color.green/(double)65535;
	*b=(double)color.blue/(double)65535;
}

static GdkColor color_tint_black(GdkColor color,gdouble t)
{
  GdkColor ncolor=color;
  ncolor.red=color.red*t;
  ncolor.green=color.green*t;
  ncolor.blue=color.blue*t;

  return ncolor;
}

/*
static GdkColor color_tint_white(GdkColor color,gdouble t)
{
  GdkColor ncolor=color;
  ncolor.red=color.red/t;
  ncolor.green=color.green/t;
  ncolor.blue=color.blue/t;

  return ncolor;
}
*/

static void gdk_get_off(GdkDrawable *window,gint* x_off,gint* y_off)
{
  GdkDrawable *real_drawable;
  
    if(GDK_IS_WINDOW(window))
    {  
      gdk_window_get_internal_paint_info (window, &real_drawable, x_off, y_off);  
    }

}

/* rename this to cge_init */
static void cge_gdk(cairo_t *xrs, GdkDrawable *window)
{
  Display *dpy;
  Drawable drawable;
  GdkDrawable *real_drawable;
  gint x_off, y_off;
  
  /* Unabstract X from GTK+ */
  if(GDK_IS_WINDOW(window))
    {  
      gdk_window_get_internal_paint_info (window, &real_drawable, &x_off, &y_off);  
      dpy = gdk_x11_drawable_get_xdisplay (real_drawable);
      drawable = gdk_x11_drawable_get_xid (real_drawable);
    }
  else
    {
      dpy = gdk_x11_drawable_get_xdisplay (window);
      drawable = gdk_x11_drawable_get_xid (window);
    }
  /* Ignore GTK+ and use Cairo for drawing. */
  cairo_set_target_drawable (xrs, dpy, drawable);

  if(GDK_IS_WINDOW(window))
    cairo_translate(xrs,-(double)x_off,-(double)y_off);
 	
}

static void cge_draw_line(cairo_t *xrs, GdkColor c, gdouble alpha, 
			     double x1, double y1,
			     double x2, double y2, 
			     gboolean line_width)
{
  gdouble r,g,b;	
  cairo_save(xrs);

  cairo_new_path(xrs);
  cairo_set_line_width(xrs, line_width);
  cairo_set_line_cap (xrs,CAIRO_LINE_CAP_ROUND);
  
  cairo_move_to(xrs, x1, y1);
  cairo_line_to(xrs, x2, y2); 
  


  color_gdk_to_rgb(c,&r,&g,&b);
  cairo_set_rgb_color(xrs, r, g, b);	
  cairo_set_alpha(xrs,alpha);
  
  cairo_close_path(xrs);
  cairo_stroke(xrs);
  cairo_restore(xrs);
}



static cairo_surface_t *cge_create_gradient(cairo_t *xrs,
					    gint ncols,
					    GdkColor *colors,
					    gdouble alpha,
					    cge_direction d,
					    gint x_off,
					    gint y_off,
					    gint width,
					    gint height)
{
  cairo_surface_t *sur;
  cairo_matrix_t *matrix;
  gint i;

  sur=cairo_surface_create_similar(cairo_current_target_surface(xrs),
				   CAIRO_FORMAT_ARGB32,
				   ncols,ncols);

  cairo_save(xrs);
  cairo_set_target_surface(xrs,sur);

  cairo_translate(xrs,(double)x_off,(double)y_off);

  for(i=0;i<ncols;i++)
    {
      gdouble r,g,b;	

      color_gdk_to_rgb(colors[i],&r,&g,&b);     
      cairo_set_rgb_color(xrs, r, g, b);	
      cairo_set_alpha(xrs,alpha);
      switch (d)
	{
	case CGE_DIRECTION_VERTICAL:
	  cairo_rectangle(xrs,0,i,ncols,1);  
	  break;
	case CGE_DIRECTION_HORIZONTAL:	  	
	  cairo_rectangle(xrs,i,0,1,ncols);	  
	  break;
	}

      cairo_fill(xrs);
      
    } 

  cairo_restore(xrs);
 
  matrix=cairo_matrix_create();
  cairo_matrix_scale(matrix,(double)(ncols-1)/width,(double)(ncols-1)/height);
  cairo_surface_set_matrix(sur,matrix);
  cairo_surface_set_filter(sur, CAIRO_FILTER_BILINEAR);
  cairo_matrix_destroy(matrix);
  return sur;
}

/* 
static void cge_draw_slider(cairo_t *xrs,
			GdkColor color,
			gdouble alpha, 
			gint x,gint y,gint w, gint h,
			gdouble line_width,
			gint line_count
			)
{
  gint i;

  cairo_save(xrs);  
  
  cairo_translate(xrs,(double)x,(double)y);
  for(i=0;i<line_count;i++)
    {
      cge_draw_line(xrs,
		       color, 
		       alpha, 
		       0, (h/line_count)*i,w, (h/line_count)*i, 
		       line_width);
    }

  cairo_restore(xrs);  
} 
*/

static void cge_draw_checked(cairo_t *xrs,
				GdkColor color,
				gdouble alpha, 
				gint x,gint y,gint w, gint h,
				gdouble line_width
				)
{

  cairo_save(xrs);  
  
  cairo_translate(xrs,(double)x,(double)y);
  cge_draw_line(xrs,
		   color, 
		   alpha, 
		   0, 0,w, h, 
		   line_width);
  cge_draw_line(xrs,
		   color, 
		   alpha, 
		   0, h,w, 0, 
		   line_width);
  
  cairo_restore(xrs);  
} 





static void cge_draw_arrow_from_rc(cairo_t *xrs,
			   GtkStyle * style,
			   GtkStateType state_type,
			   const gchar *detail,
			   GdkColor c,
			   gdouble alpha, 
			   gint x,gint y,gint w,gint h,
			   gboolean fill,
			   gdouble line_width, 
			   GtkArrowType at)
{
  gdouble r,g,b;
  cairo_save(xrs);
  cairo_new_path(xrs);
  cairo_set_line_width(xrs, line_width);
  cairo_set_line_join (xrs, CAIRO_LINE_JOIN_ROUND);
  gdouble rt;

  switch(THEME_DATA(style)->arrow_style)
    {      
    case CGE_ARROW_NORMAL:
      rt=h/4 - h/10;
      break;
    case CGE_ARROW_SIMPLE:
      rt=h/4;
      break;
    default:
      break;
    }


  switch(at)
    {
    case GTK_ARROW_UP:
      cairo_translate(xrs,(double)x,(double)y);
      cairo_move_to(xrs, (double)w/2, rt);
      break;
    case GTK_ARROW_DOWN:
      cairo_translate(xrs,(double)x,(double)y);
      cairo_move_to(xrs, (double)w/2, h-rt);
      cairo_rotate (xrs, M_PI);          
      break;
    case GTK_ARROW_LEFT:
      cairo_translate(xrs,(double)x,(double)y);
      cairo_move_to(xrs, rt, (double)h/2);
      cairo_rotate (xrs, -M_PI_2);    
      break;
    case GTK_ARROW_RIGHT:
      cairo_translate(xrs,(double)x,(double)y);
      cairo_move_to(xrs, (double)w-rt, (double)h/2);
      cairo_rotate (xrs, M_PI_2);    
      break;

    }  

  switch(THEME_DATA(style)->arrow_style)
    {      
    case CGE_ARROW_NORMAL:
      cairo_rel_line_to(xrs,  (double) w/2,   (double)h/2);
      cairo_rel_line_to(xrs, -(double)w/2,  (double)h/5);
      cairo_rel_line_to(xrs, -(double)w/2,  -(double)h/5);
      break;
    case CGE_ARROW_SIMPLE:
      cairo_rel_line_to(xrs,  (double) w/2,   (double)h/2);      
      cairo_rel_line_to(xrs, -(double)w,0);
      break;
    default:
      break;
    }
  
  //printf("%i\n",THEME_DATA(style)->arrow_style);

  color_gdk_to_rgb(c,&r,&g,&b);
  cairo_set_rgb_color(xrs, r, g, b);	
  cairo_set_alpha(xrs,alpha);
  cairo_close_path(xrs);    



  if(fill)
    cairo_fill(xrs);
  else	
    cairo_stroke(xrs);
  cairo_restore(xrs);
}

static void cge_draw_rectangle(cairo_t *xrs,
			   GdkColor c,
				  
			   gdouble alpha,
			   gint x,
			   gint y,
			   gint w, 
			   gint h)
{
  gdouble r,g,b;	

  cairo_save(xrs);
  cairo_translate(xrs,(double)x,(double)y);
  cairo_rectangle(xrs,0,0,(double)w,(double)h);
  color_gdk_to_rgb(c,&r,&g,&b);  
  cairo_set_alpha(xrs,alpha);
  cairo_set_rgb_color(xrs, r, g, b);	   
  cairo_fill(xrs);
  cairo_restore(xrs);

  
}


/* FIXME: need to be merged with cge_draw_box 
   adding a gboolean beveled parameter 
   whith that button can be beveled and rounded
*/
static void cge_draw_beveled_box(cairo_t *xrs,
		     GdkColor c1,
		     GdkColor c2,
		     gdouble alpha,
		     gint x,
		     gint y,
		     gint w, 
		     gint h, 
		     gdouble line_width)
{
  gdouble r,g,b;	

  cairo_save(xrs);

  /*  x += line_width;
  y += line_width;
  w -= line_width*2;
  h -= line_width*2; 
  */
  cairo_save(xrs);
  cairo_new_path(xrs);
  cairo_set_line_width(xrs, line_width/2);  
  cairo_translate(xrs,(double)x,(double)y);
  cairo_move_to(xrs, 0,(double)h);  
  cairo_rel_line_to(xrs,  0,   -(double)h);
  cairo_rel_line_to(xrs,  (double)w,   0);

  color_gdk_to_rgb(c2,&r,&g,&b);
  cairo_set_rgb_color(xrs, r, g, b);	
  cairo_set_alpha(xrs,alpha);

  /*
  cairo_set_rgb_color(xrs, 0, 0, 0);	   
  cairo_set_alpha(xrs,0.6);
  */

  cairo_stroke(xrs);  
  cairo_restore(xrs);

  cairo_save(xrs);
  cairo_new_path(xrs);
  cairo_set_line_width(xrs, line_width);  
  cairo_translate(xrs,(double)x,(double)y);
  cairo_move_to(xrs, line_width/2,(double)h+line_width/2);  
  cairo_rel_line_to(xrs,  0,   -(double)h);
  cairo_rel_line_to(xrs,  (double)w,   0);

  color_gdk_to_rgb(c1,&r,&g,&b);
  cairo_set_rgb_color(xrs, r, g, b);	
  cairo_set_alpha(xrs,alpha);

  /* 
  cairo_set_rgb_color(xrs, 1, 1, 1);	   
  cairo_set_alpha(xrs,0.6);
  */

  cairo_stroke(xrs);  
  cairo_restore(xrs);

  cairo_save(xrs);
  cairo_new_path(xrs);
  cairo_set_line_width(xrs, line_width);  
  cairo_translate(xrs,(double)x,(double)y);
  cairo_move_to(xrs, 0,(double)h);  
  cairo_rel_line_to(xrs,  (double)w,   0);
  cairo_rel_line_to(xrs,  0,   -(double)h);

  color_gdk_to_rgb(c2,&r,&g,&b);
  cairo_set_rgb_color(xrs, r, g, b);	
  cairo_set_alpha(xrs,alpha);

  /*
  cairo_set_rgb_color(xrs, 0, 0, 0);	   
  cairo_set_alpha(xrs,0.6);
  */
  cairo_stroke(xrs);  
  cairo_restore(xrs);


  cairo_restore(xrs);
}
			   


static void cge_draw_box(cairo_t *xrs,
		     GdkColor c,
		     cairo_surface_t *pattern,
		     gdouble alpha,
		     gdouble x,
		     gdouble y,
		     gdouble w, 
		     gdouble h, 
		     gdouble round_size, 
		     gboolean fill,
		     gdouble line_width)
{
  gdouble r,g,b;	

  cairo_save(xrs);
       
  cairo_new_path(xrs);
  cairo_set_line_width(xrs, line_width);
  cairo_set_line_join (xrs, CAIRO_LINE_JOIN_ROUND);

  if(!fill){
    x += line_width/2;
    y += line_width/2;
    w -= line_width;
    h -= line_width; 
  }


  cairo_translate(xrs,(double)x,(double)y);
  cairo_move_to(xrs, 0,(double)round_size);
  cairo_rel_curve_to (xrs,
		      0, -(double)round_size,
		      (double)round_size/2, -(double)round_size,
		      (double)round_size, -(double)round_size);

  cairo_rel_line_to(xrs,  (double) w-(double)2*round_size,   0);

  cairo_rel_curve_to (xrs,
		      (double)round_size,0,
		      (double)round_size,(double)round_size/2,
		      (double)round_size,(double)round_size);

  cairo_rel_line_to(xrs,   0,  (double)h-(double)2*round_size);
  cairo_rel_curve_to (xrs,
		      0, (double)round_size,
		      -(double)round_size/2, (double)round_size,
		      -(double)round_size, (double)round_size);

  cairo_rel_line_to(xrs, -(double)w+(double)2*round_size,   0); 
  cairo_rel_curve_to (xrs,
		      -(double)round_size,0,
		      -(double)round_size,-(double)round_size/2,
		      -(double)round_size,-(double)round_size);

  color_gdk_to_rgb(c,&r,&g,&b);
  
  cairo_set_rgb_color(xrs, r, g, b);	   
  if(pattern!=NULL) 
    cairo_set_pattern(xrs,pattern);

  cairo_set_alpha(xrs,alpha);

  
  cairo_close_path(xrs);



  if(fill)
    cairo_fill(xrs);
  else	
    cairo_stroke(xrs);
  cairo_restore(xrs);
}


static void cge_draw_circle(cairo_t *xrs,
		     GdkColor c,
		     cairo_surface_t *pattern,
		     gdouble alpha,
		     gint x,
		     gint y,
		     gdouble round_size, 
		     gboolean fill,
		     gdouble line_width)
{
  gdouble r,g,b;	

  cairo_save(xrs);

  

  if(!fill)
    round_size -=line_width/2;
  
  cairo_new_path(xrs);
  cairo_set_line_width(xrs, line_width);
  cairo_set_line_join (xrs, CAIRO_LINE_JOIN_ROUND);
  //cairo_set_tolerance (xrs, 5);


  cairo_translate(xrs,(double)x,(double)y);
  cairo_move_to(xrs, 0,(double)round_size);
  cairo_rel_curve_to (xrs,
		      0, -(double)round_size,
		      (double)round_size/2, -(double)round_size,
		      (double)round_size, -(double)round_size);

  cairo_rel_curve_to (xrs,
		      (double)round_size,0,
		      (double)round_size,(double)round_size/2,
		      (double)round_size,(double)round_size);

  cairo_rel_curve_to (xrs,
		      0, (double)round_size,
		      -(double)round_size/2, (double)round_size,
		      -(double)round_size, (double)round_size);

  cairo_rel_curve_to (xrs,
		      -(double)round_size,0,
		      -(double)round_size,-(double)round_size/2,
		      -(double)round_size,-(double)round_size);

  color_gdk_to_rgb(c,&r,&g,&b);
  
  cairo_set_alpha(xrs,alpha);

  cairo_set_rgb_color(xrs, r, g, b);	   
  if(pattern!=NULL) 
    cairo_set_pattern(xrs,pattern);



  
  cairo_close_path(xrs);



  if(fill)
    cairo_fill(xrs);
  else	
    cairo_stroke(xrs);
  cairo_restore(xrs);
}


/* FIXME : work only with top gap*/
static void cge_draw_box_gap(cairo_t *xrs,
			 GdkColor c,
			 cairo_surface_t *pattern,
			 gdouble alpha,
			 gdouble x,
			 gdouble y,
			 gdouble w, 
			 gdouble h, 
			 gdouble round_size, 
			 gint gap_x,
			 gint gap_width,
			 gboolean fill,
			 gdouble line_width)
{
  gdouble r,g,b;	

  cairo_save(xrs);
  cairo_new_path(xrs);
  cairo_set_line_width(xrs, line_width);
  cairo_set_line_join (xrs, CAIRO_LINE_JOIN_ROUND);
  cairo_set_line_cap (xrs,CAIRO_LINE_CAP_SQUARE);

  if (!fill) {
      x += line_width/2;
      y += line_width/2;
      w -= line_width;
      h -= line_width;
  } 

  cairo_translate(xrs,(double)x,(double)y);
  cairo_move_to(xrs, 0,0);
  cairo_rel_line_to(xrs,  (double)gap_x,   0);

  cairo_move_to(xrs, (double)gap_x+(double)gap_width-line_width,0);
  cairo_rel_line_to(xrs,  (double) w-(double)round_size-(double)gap_width-(double)gap_x+line_width,   0);


  cairo_rel_curve_to (xrs,
		      (double)round_size,0,
		      (double)round_size,(double)round_size/2,
		      (double)round_size,(double)round_size);

  cairo_rel_line_to(xrs,   0,  (double)h-(double)2*round_size);
  cairo_rel_curve_to (xrs,
		      0, (double)round_size,
		      -(double)round_size/2, (double)round_size,
		      -(double)round_size, (double)round_size);

  cairo_rel_line_to(xrs, -(double)w+(double)2*round_size,   0); 
  cairo_rel_curve_to (xrs,
		      -(double)round_size,0,
		      -(double)round_size,-(double)round_size/2,
		      -(double)round_size,-(double)round_size);

  cairo_rel_line_to(xrs, 0,   -(double)h); 


  color_gdk_to_rgb(c,&r,&g,&b);
  
  cairo_set_alpha(xrs,alpha);

  cairo_set_rgb_color(xrs, r, g, b);	   

  if(pattern!=NULL) 
    cairo_set_pattern(xrs,pattern);
  

  if(fill)
    cairo_fill(xrs);
  else	
    cairo_stroke(xrs);
  cairo_restore(xrs);
}


/* FIXME : work only with top notebook tab*/
static void cge_draw_notebook_tab(cairo_t *xrs,
				   GdkColor c,
				   cairo_surface_t *pattern,
				   gdouble alpha,
				   gdouble x,
				   gdouble y,
				   gdouble w, 
				   gdouble h, 
				   gdouble round_size, 
				   gboolean fill,
				   gdouble line_width)
{

  gdouble r,g,b;	

  cairo_save(xrs);
  cairo_new_path(xrs);
  cairo_set_line_width(xrs, line_width);
  cairo_set_line_join (xrs, CAIRO_LINE_JOIN_ROUND);
  cairo_set_line_cap (xrs,CAIRO_LINE_CAP_SQUARE);

  if (!fill) 
    {
      x += line_width/2;
      y += line_width/2;
      w -= line_width;
      h -= line_width;
    } 

  cairo_translate(xrs,(double)x,(double)y);
  cairo_move_to(xrs, 0,(double)h);
  cairo_rel_line_to(xrs,  0,  -(double)h+(double)round_size );
  cairo_rel_curve_to (xrs,
		      0, -(double)round_size,
		      (double)round_size/2, -(double)round_size,
		      (double)round_size, -(double)round_size);

  cairo_rel_line_to(xrs,  (double) w-(double)2*round_size,   0);
  cairo_rel_curve_to (xrs,
		      (double)round_size,0,
		      (double)round_size,(double)round_size/2,
		      (double)round_size,(double)round_size);

  cairo_rel_line_to(xrs,   0,  (double)h-(double)round_size);


  color_gdk_to_rgb(c,&r,&g,&b);
  
  cairo_set_alpha(xrs,alpha);

  cairo_set_rgb_color(xrs, r, g, b);	   
  if(pattern!=NULL) 
    cairo_set_pattern(xrs,pattern);


  if(fill)
    cairo_fill(xrs);
  else	
    cairo_stroke(xrs);
  cairo_restore(xrs);
}

cairo_surface_t *cge_create_gradient_from_rc(cairo_t *xrs,
						GtkStyle * style,
						GtkStateType state_type,
						const gchar *detail,
						gint x_off,
						gint y_off,
						gint width,
						gint height)
{
  cairo_surface_t *gradient;
  gint i;
  cge_style *cstyle;
  cstyle=cge_get_style(style,detail);
  GdkColor colors[cstyle->fill_style.gradient_colors_count[state_type]];


  //printf("gradient detail: %s\n",detail);

  if(cstyle->fill_style.fill_mode==CGE_FILL_SOLID)
    return NULL;
  
  if(cstyle->fill_style.gradient_colors_default[state_type]==TRUE)
    {
      switch(state_type)
	{
	case GTK_STATE_PRELIGHT:
	  colors[0]=style->bg[GTK_STATE_PRELIGHT];
	  colors[1]=style->bg[GTK_STATE_SELECTED];
	  
	  break;
	default:	  
	  colors[0]=style->bg[state_type];
	  colors[1]=color_tint_black(style->bg[state_type],0.96);      
	}
    }
  else
    {
      for(i=0;i<cstyle->fill_style.gradient_colors_count[state_type];i++)
	{
	  GdkColor tmp=colors[i]=cstyle->fill_style.gradient_colors[state_type][i];
	  
	  colors[i]=tmp;
	}
    }
  
  
  
  
  if(cstyle->fill_style.fill_mode==CGE_FILL_GRADIENT)
    gradient=cge_create_gradient(xrs,
				 cstyle->fill_style.gradient_colors_count[state_type],
				 colors,
				 cstyle->fill_style.alpha,
				 cstyle->fill_style.gradient_direction,
				 x_off,y_off,width,height);
  
  return gradient;
}


static void cge_draw_filled_box_from_rc(cairo_t *xrs,
				    GtkStyle * style,
				    GtkStateType state_type,
				    const gchar *detail,
				    gint x_off,
				    gint y_off,
				    gint x,
				    gint y,
				    gint width,
				    gint height)

{
  cairo_surface_t *gradient;
  GdkColor color;
  cge_style *cstyle;

  
  cstyle=cge_get_style(style,detail);

  gradient=cge_create_gradient_from_rc(xrs, style,state_type,detail,x_off,y_off,width,height); 


  if(DETAIL("entry_bg") || DETAIL("cell_odd") || DETAIL("cell_even")) {
    color=style->base[state_type];
    
    cge_draw_rectangle(xrs,color,1,x,y,width,height);
  }
  else if(DETAIL("background"))
    {
	  color=style->bg[state_type];
      	  cge_draw_box(xrs,
		       color,
		       gradient,
		       1,
		       x,
		       y,
		       width,
		       height,
		       0,
		       TRUE,0);
    }
  else
    {
      if(cstyle->line_style.line_mode==CGE_LINE_BEVEL)
	{
	  cairo_save(xrs);
	  cairo_translate(xrs,(double)x,(double)y);
	  cairo_rectangle(xrs,0,0,(double)width,(double)height);
	  cairo_set_pattern(xrs,gradient);
	  cairo_fill(xrs);
	  cairo_restore(xrs);
	}
      else
	{
	  color=style->bg[state_type];

	  
	  cge_draw_box(xrs,
		       color,
		       gradient,
		       cstyle->fill_style.alpha,
		       x,
		       y,
		       width,
		       height,
		       cstyle->fill_style.round_width,
		       TRUE,0);
	}    
    }
  

  if(gradient)
    cairo_surface_destroy(gradient);


}



static void cge_draw_unfilled_box_from_rc(cairo_t *xrs,
				      GtkStyle * style,
				      GtkStateType state_type,
				      const gchar *detail,
				      gint x,
				      gint y,
				      gint width,
				      gint height)

{
  GdkColor color;
  cge_style *cstyle;


  cairo_save(xrs);

  cstyle=cge_get_style(style,detail);

  if(cstyle->line_style.color_default[state_type]==TRUE)
    color=style->dark[state_type];
  else
    color=cstyle->line_style.color[state_type];

  if(cstyle->line_style.line_mode==CGE_LINE_DASHED)
    {
      // for now, static dash
      double d[2]={4,4};
      cairo_set_dash (xrs, d, 2, 0);
    }
  
  if(cstyle->line_style.line_mode==CGE_LINE_BEVEL)
    {
      GdkColor light_color;
      GdkColor dark_color;
      switch(state_type)
	{
	case GTK_STATE_NORMAL:
	  light_color=style->light[state_type];
	  dark_color=style->dark[state_type];
	  break;
	case GTK_STATE_PRELIGHT:
	  light_color=style->bg[state_type];
	  dark_color=style->dark[state_type];
	  break;
	case GTK_STATE_ACTIVE:
	  light_color=style->dark[state_type];
	  dark_color=style->light[state_type];
	  break;
	      
	default:	  
	  light_color=style->light[state_type];
	  dark_color=style->dark[state_type];
	  break;
	}
      
      if(DETAIL("scrolled_window"))
	{
	  light_color=style->dark[state_type];
	  dark_color=style->light[state_type];
	}
      cge_draw_beveled_box(xrs,			  
			   light_color,
			   dark_color,
			   cstyle->line_style.alpha,
			   x,
			   y,
			   width,
			   height,
			   cstyle->line_style.line_width);
    }
  else if(cstyle->line_style.line_mode==CGE_LINE_SOLID)
    cge_draw_box(xrs,
		 color,
		 NULL,
		 cstyle->line_style.alpha,
		 x,
		 y,
		 width,
		 height,
		 cstyle->line_style.round_width,
		 FALSE,
		 cstyle->line_style.line_width);
      

  cairo_restore(xrs);
}

static void cge_draw_line_from_rc(cairo_t *xrs,
			      GtkStyle * style,
			      GtkStateType state_type,
			      const gchar *detail,
			      gdouble x1,
			      gdouble y1,
			      gdouble x2,
			      gdouble y2)

{
  GdkColor color;
  cge_style *cstyle;

  cstyle=cge_get_style(style,detail);

  if(cstyle->line_style.color_default[state_type]==TRUE)
    color=style->dark[state_type];
  else
    color=cstyle->line_style.color[state_type];

  cairo_save(xrs);
      
  if(cstyle->line_style.line_mode==CGE_LINE_DASHED)
    {
      // for now, static dash
      double d[2]={4,4};
      cairo_set_dash (xrs, d, 2, 0);
    }

  cge_draw_line(xrs,
		   color,
		   1,
		   x1,
		   y1,
		   x2,
		   y2,
		   cstyle->line_style.line_width);

  cairo_restore(xrs);
}

static void
draw_hline (GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    const gchar * detail, gint x1, gint x2, gint y)
{
  cairo_t *xrs;
  xrs=cairo_create();			
  cge_gdk(xrs,window);

  cge_draw_line_from_rc(xrs,
			   style,
			   state_type,
			   detail,
			   x1,
			   y+0.5,
			   x2,
			   y+0.5);

  cairo_destroy(xrs);
  
}

static void
draw_vline (GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    const gchar * detail, gint y1, gint y2, gint x)
{
  cairo_t *xrs;
  xrs=cairo_create();			
  cge_gdk(xrs,window);

  cge_draw_line_from_rc(xrs,
			   style,
			   state_type,
			   detail,
			   x+0.5,
			   y1,
			   x+0.5,
			   y2);

  cairo_destroy(xrs);


}



static void draw_tab (GtkStyle *style,
		      GdkWindow *window,
		      GtkStateType state_type,
		      GtkShadowType shadow_type,
		      GdkRectangle *area,
		      GtkWidget *widget,
		      const gchar *detail,
		      gint x,
		      gint y,
		      gint width,
		      gint height)
{
  cairo_t *xrs;
  xrs=cairo_create();			
  cge_gdk(xrs,window);

  cge_draw_arrow_from_rc(xrs,
			 style,
			 state_type,
			 detail,
			 style->text[state_type],
			 1,
			 x+1,y,width-2,(height-2)/2,
			 TRUE,1.5,GTK_ARROW_UP);
  cge_draw_arrow_from_rc(xrs,
			 style,
			 state_type,
			 detail,
			 style->text[state_type],
			 1,
			 x+1,y+3+(height-2)/2,width-2,(height-2)/2,
			 TRUE,1.5,GTK_ARROW_DOWN);

  cairo_destroy(xrs);
}

static void
draw_shadow (GtkStyle * style, 
	     GdkWindow * window,
	     GtkStateType state_type, 
	     GtkShadowType shadow_type,
	     GdkRectangle * area,
	     GtkWidget * widget, 
	     const gchar * detail,
	     gint x, 
	     gint y, 
	     gint width, 
	     gint height)

{
  cairo_t *xrs;
  cge_style *cstyle;

  cstyle=cge_get_style(style,detail);

  xrs=cairo_create();			

  cge_gdk(xrs,window);
  
  if(DETAIL("entry"))
    cge_draw_box(xrs,
		    style->dark[state_type],
		    NULL,
		    1,
		    x,y,width,height,
		    0,
		    FALSE,
		    cstyle->line_style.line_width);

  //cge_draw_rectangle(xrs,style->bg[state_type],1,x,y,width,height);

  else
  
  cge_draw_unfilled_box_from_rc(xrs,
				   style,
				   state_type,
				   detail,
				   x,
				   y,
				   width,
				   height);
    
  
  cairo_destroy(xrs);
  
}

static void draw_shadow_gap (GtkStyle               *style,
			     GdkWindow              *window,
			     GtkStateType            state_type,
			     GtkShadowType           shadow_type,
			     GdkRectangle           *area,
			     GtkWidget              *widget,
			     const gchar            *detail,
			     gint                    x,
			     gint                    y,
			     gint                    width,
			     gint                    height,
			     GtkPositionType         gap_side,
			     gint                    gap_x,
			     gint                    gap_width)
{
  cairo_t *xrs;
  gint x_off=0,y_off=0;
  GdkColor color;
  cairo_surface_t *gradient;
  cge_style *cstyle;

  cstyle=cge_get_style(style,detail);

  xrs=cairo_create();			

  cge_gdk(xrs,window);

  gdk_get_off(window,&x_off,&y_off);

  gradient=cge_create_gradient_from_rc(xrs, 
				       style,
				       state_type,
				       detail,
				       x_off,y_off,width,height); 
  
  if(cstyle->line_style.color_default[state_type]==TRUE)
    color=style->dark[state_type];
  else
    color=cstyle->line_style.color[state_type];

  cge_draw_box_gap(xrs,
		   style->bg[state_type],
		   NULL,
		   cstyle->fill_style.alpha,
		   x,y,width,height,
		   cstyle->fill_style.round_width,
		   gap_x,gap_width,
		   TRUE,0);

  cge_draw_box_gap(xrs,
		      color,
		      NULL,
		      cstyle->line_style.alpha,
		      x,y,width,height,
		      cstyle->line_style.round_width,
		      gap_x,gap_width,
		      FALSE,
		      cstyle->line_style.line_width);

  if(gradient)
    cairo_surface_destroy(gradient);
  
  cairo_destroy(xrs);
}

static void
draw_arrow (GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GtkShadowType shadow_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    const gchar * detail,
	    GtkArrowType arrow_type,
	    gint fill, gint x, gint y, gint width, gint height)

{
  cairo_t *xrs;
  xrs=cairo_create();			
  cge_gdk(xrs,window);

  cge_draw_arrow_from_rc(xrs,
		 style,
		 state_type,
		 detail,
		 style->text[state_type],
		 1,
		 x,y,width,height,
		 TRUE,1.5,arrow_type);
  
  cairo_destroy(xrs);
}

static void
draw_focus (GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    const gchar * detail, gint x, gint y, gint width, gint height)
{
  cairo_t *xrs;
  cge_style *cstyle;

  cstyle=cge_get_style(style,detail);

  xrs=cairo_create();		
	
  cge_gdk(xrs,window);

  if(!GTK_IS_NOTEBOOK(widget))
    cge_draw_box(xrs,
		    style->dark[state_type],
		    NULL,
		    cstyle->line_style.alpha,
		    x+2,y+2,width-4,height-4,
		    cstyle->line_style.round_width,
		    FALSE,
		    cstyle->line_style.line_width/2);
 

  cairo_destroy(xrs);

}


static void
draw_box (GtkStyle * style,
	  GdkWindow * window,
	  GtkStateType state_type,
	  GtkShadowType shadow_type,
	  GdkRectangle * area,
	  GtkWidget * widget,
	  const gchar * detail, gint x, gint y, gint width, gint height)
{ 
  cairo_t *xrs;
  gint x_off=0,y_off=0;

  xrs=cairo_create();
  
  cge_gdk(xrs,window);
  cairo_save(xrs);
  gdk_get_off(window,&x_off,&y_off);
  //cairo_translate(xrs,-(double)x_off,-(double)y_off);  
  
  //printf("detail(draw_box) : %s\n",detail);

 
  if(DETAIL("trough"))
    cge_draw_rectangle(xrs,style->bg[GTK_STATE_NORMAL],1,x,y,width,height); 
  

  if(DETAIL("menu")
     ) {
  }
  else
    {
      cge_draw_filled_box_from_rc(xrs,
				     style,
				     state_type,
				     detail,
				     x_off,
				     y_off,
				     x,
				     y,
				     width,
				     height);
      
      cge_draw_unfilled_box_from_rc(xrs,
				    style,
				    state_type,
				    detail,
				    x,
				    y,
				    width,
				    height);
      
    }
  cairo_restore(xrs);     
  cairo_destroy(xrs);
}

static void draw_box_gap (GtkStyle *style,
			  GdkWindow *window,
			  GtkStateType state_type,
			  GtkShadowType shadow_type,
			  GdkRectangle *area,
			  GtkWidget *widget,
			  const gchar *detail,
			  gint x,
			  gint y,
			  gint width,
			  gint height,
			  GtkPositionType gap_side,
			  gint gap_x,
			  gint gap_width)
{
  cairo_t *xrs;
  gint x_off=0,y_off=0;
  GdkColor color;
  cairo_surface_t *gradient;
  cge_style *cstyle;

  cstyle=cge_get_style(style,detail);

  xrs=cairo_create();			

  cge_gdk(xrs,window);

  gdk_get_off(window,&x_off,&y_off);

  gradient=cge_create_gradient_from_rc(xrs,
				       style,
				       state_type,
				       detail,
				       x_off,y_off,width,height); 

  //printf("detail : %s\n",detail);
  //cge_draw_rectangle(xrs,style->bg[state_type],1,x,y,width,height); 

  

  if(cstyle->line_style.color_default[state_type]==TRUE)
    color=style->dark[state_type];
  else
    color=cstyle->line_style.color[state_type];

  cge_draw_box_gap(xrs,
		   style->bg[state_type],
		   NULL,
		   cstyle->fill_style.alpha,
		   x,y,width,height,
		   cstyle->fill_style.round_width,
		   gap_x,gap_width,
		   TRUE,0);

  cge_draw_box_gap(xrs,
		   color,
		   NULL,
		   cstyle->line_style.alpha,
		   x,y,width,height,
		   cstyle->line_style.round_width,
		   gap_x,gap_width,
		   FALSE,
		   cstyle->line_style.line_width);
  
  if(gradient)
    cairo_surface_destroy(gradient);
  
  cairo_destroy(xrs);
}

static void
draw_flat_box (GtkStyle * style,
	       GdkWindow * window,
	       GtkStateType state_type,
	       GtkShadowType shadow_type,
	       GdkRectangle * area,
	       GtkWidget * widget,
	       const gchar * detail, gint x, gint y, gint width, gint height)
{
  cairo_t *xrs;
  gint x_off=0,y_off=0;
  xrs=cairo_create();			
  cge_gdk(xrs,window);

  gdk_get_off(window,&x_off,&y_off);
  
  cge_draw_rectangle(xrs,style->bg[state_type],1,x,y,width,height);
  //printf("detail(draw_flat_box) : %s\n",detail);


  cge_draw_filled_box_from_rc(xrs,
			      style,
			      state_type,
			      detail,
			      x_off,
			      y_off,
			      x,
			      y,
			      width,
			      height);

  
    

  cairo_destroy(xrs);

}

static void
draw_check (GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GtkShadowType shadow_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    const gchar * detail, 
	    gint x, gint y, gint width, gint height)
{
  cairo_t *xrs;
  cge_style *cstyle;

  cstyle=cge_get_style(style,detail);

  xrs=cairo_create();			
  cge_gdk(xrs,window);

  if (DETAIL ("check")) {
    cge_draw_box(xrs,
		    style->bg[state_type],
		    NULL,
		    1,
		    x,y,width,height,
		    cstyle->fill_style.round_width,
		    TRUE,0);    
      if (shadow_type == GTK_SHADOW_IN) {
	// here vectored check
	cge_draw_checked(xrs,
			    style->text[state_type],			   
			    1,
			    x+2,y+2,width-4,height-4,
			    cstyle->line_style.line_width);
      }
  }
  else
    {
      cge_draw_box(xrs,
		      style->base[state_type],
		      NULL,
		      1,
		      x,y,width,height,
		      cstyle->fill_style.round_width,
		      TRUE,0);    
      
      cge_draw_unfilled_box_from_rc(xrs,
				       style,
				       state_type,
				       detail,
				       x,
				       y,
				       width,
				       height);
  
      if (shadow_type == GTK_SHADOW_IN) {
	// here vectored check
	cge_draw_checked(xrs,
			    style->text[state_type],			   
			    1,
			    x+2,y+2,width-4,height-4,
			    cstyle->line_style.line_width);
		
      }
  }

  

  cairo_destroy(xrs);
}

// FIXME : make a true circle draw function
static void
draw_option (GtkStyle * style,
	     GdkWindow * window,
	     GtkStateType state_type,
	     GtkShadowType shadow_type,
	     GdkRectangle * area,
	     GtkWidget * widget,
	     const gchar * detail, gint x, gint y, gint width, gint height)
{

  cairo_t *xrs;
  cge_style *cstyle;

  cstyle=cge_get_style(style,detail);

  xrs=cairo_create();			
  cge_gdk(xrs,window);
  
  //cge_draw_rectangle(xrs,style->bg[state_type],1,x,y,width,height); 

  if (DETAIL ("option")) {
    cge_draw_circle(xrs,
		    style->bg[state_type],
		    NULL,
		    1,
		    x,y,
		    width/2,
		    TRUE,0);    
      if (shadow_type == GTK_SHADOW_IN) {
	// here vectored option
	cge_draw_circle(xrs,
			style->fg[state_type],
			NULL,
			1,
			x+1,y+1,
			width/2-2,
			TRUE,0);    
      }
  }
  else
    {
      cge_draw_circle(xrs,
		      style->base[state_type],
		      NULL,
		      1,
		      x,y,
		      width/2,
		      TRUE,0);    
      cge_draw_circle(xrs,
		      style->dark[state_type],
		      NULL,
		      1,
		      x,y,
		      width/2,
		      FALSE,
		      cstyle->line_style.line_width);

  
      if (shadow_type == GTK_SHADOW_IN) {
	// here vectored option
	cge_draw_circle(xrs,
			style->text[state_type],
			NULL,
			1,
			x+2,y+2,
			(width)/2 - 2,
			TRUE,0);    
      }
  }

  

  cairo_destroy(xrs);

}


/* tab */
static void
draw_extension (GtkStyle * style,
		GdkWindow * window,
		GtkStateType state_type,
		GtkShadowType shadow_type,
		GdkRectangle * area,
		GtkWidget * widget,
		const gchar * detail,
		gint x, gint y, gint width, gint height,
		GtkPositionType gap_side)
{
  cairo_t *xrs;
  cairo_surface_t *gradient;
  GdkColor color;
  gint x_off,y_off;
  cge_style *cstyle;

  cstyle=cge_get_style(style,detail);

  xrs=cairo_create();			
  cge_gdk(xrs,window);

  gdk_get_off(window,&x_off,&y_off);

  gradient=cge_create_gradient_from_rc(xrs,
					  style,
					  state_type,
					  detail,
					  x_off,y_off,width,height); 

  //cge_draw_rectangle(xrs,style->bg[GTK_STATE_NORMAL],1,x,y,width,height); 
	
  if(cstyle->line_style.color_default[state_type]==TRUE)
    color=style->dark[state_type];
  else
    color=cstyle->line_style.color[state_type];

  cge_draw_notebook_tab(xrs,
			   style->bg[state_type],
			   gradient,
			   cstyle->fill_style.alpha,
			   x,y,width,height,
			   cstyle->fill_style.round_width,
			   TRUE,0);
  cge_draw_notebook_tab(xrs,
			   color,
			   NULL,
			   cstyle->line_style.alpha,
			   x,y,width,height,
			   cstyle->line_style.round_width,
			   FALSE,
			   cstyle->line_style.line_width);

  if(gradient)
    cairo_surface_destroy(gradient);
  
  cairo_destroy(xrs);

}

static void
draw_slider (GtkStyle * style,
	     GdkWindow * window,
	     GtkStateType state_type,
	     GtkShadowType shadow_type,
	     GdkRectangle * area,
	     GtkWidget * widget,
	     const gchar * detail,
	     gint x, gint y, gint width, gint height,
	     GtkOrientation orientation)
{
  cairo_t *xrs;
  gint x_off=0,y_off=0;

  xrs=cairo_create();			
  cge_gdk(xrs,window);

  gdk_get_off(window,&x_off,&y_off);

  cge_draw_filled_box_from_rc(xrs,
			      style,
			      state_type,
			      detail,
			      x_off,
			      y_off,
			      x,
			      y,
			      width,
			      height);

  cge_draw_unfilled_box_from_rc(xrs,
				style,
				state_type,
				detail,
				x,
				y,
				width,
				height);
  

  /*
  cge_draw_slider(xrs,
		     style->dark[state_type],
		     1,
		     x+4,y+height/2 - height/8,width-8,height/4,
		     THEME_DATA(style)->line_style.line_width,3);
  */
  cairo_destroy(xrs);

}

static void
draw_handle (GtkStyle * style,
	     GdkWindow * window,
	     GtkStateType state_type,
	     GtkShadowType shadow_type,
	     GdkRectangle * area,
	     GtkWidget * widget,
	     const gchar * detail,
	     gint x, gint y, gint width, gint height,
	     GtkOrientation orientation)
{
  cairo_t *xrs;
  gint x_off=0,y_off=0;

  xrs=cairo_create();			
  cge_gdk(xrs,window);

  gdk_get_off(window,&x_off,&y_off);

  //cge_draw_rectangle(xrs,style->bg[GTK_STATE_NORMAL],1,x,y,width,height); 

  cge_draw_filled_box_from_rc(xrs,
				 style,
				 state_type,
				 detail,
				 x_off,
				 y_off,
				 x,
				 y,
				 width,
				 height);

  cge_draw_unfilled_box_from_rc(xrs,
				   style,
				   state_type,
				   detail,
				   x,
				   y,
				   width,
				   height);

  
  cairo_destroy(xrs);

}

static void
draw_resize_grip (GtkStyle * style, 
		  GdkWindow * window,
		  GtkStateType state_type, 
		  GdkRectangle * area,
		  GtkWidget * widget, const gchar * detail,
		  GdkWindowEdge edge, gint x, gint y, gint width, gint height)
{
  cairo_t *xrs;
  gint x_off=0,y_off=0;

  xrs=cairo_create();			
  cge_gdk(xrs,window);

  gdk_get_off(window,&x_off,&y_off);

  //cge_draw_rectangle(xrs,style->bg[GTK_STATE_NORMAL],1,x,y,width,height); 

  cge_draw_filled_box_from_rc(xrs,
				 style,
				 state_type,
				 detail,
				 x_off,
				 y_off,
				 x+1,
				 y+1,
				 width-2,
				 height-2);

  cge_draw_unfilled_box_from_rc(xrs,
				   style,
				   state_type,
				   detail,
				   x+1,
				   y+1,
				   width-2,
				   height-2);

  cairo_destroy(xrs);

}

static void set_background (GtkStyle      *style,
			    GdkWindow     *window,
			    GtkStateType   state_type)
{
  gint x,y,w,h,d;
  gint x_off=0,y_off=0;
  cairo_t *xrs;
  GdkPixmap *pm;
  GdkRectangle rect;

  gdk_window_get_geometry         (window,
				   &x,
				   &y,
				   &w,
				   &h,
				   &d);
  rect.x=x;
  rect.y=y;
  rect.width=w;
  rect.height=h;
 
  xrs=cairo_create();			

  

  /*gdk_window_set_background       (window,
				  &style->bg[state_type]);

  */

  pm = gdk_pixmap_new(window,w,h,d);
  cge_gdk(xrs,pm);

  gdk_window_begin_paint_rect     (window,
				   &rect);
  gdk_get_off(window,&x_off,&y_off);
  gdk_window_end_paint            (window);

  cairo_translate(xrs,-(double)x_off,-(double)y_off);
  /*  cge_draw_rectangle(xrs,
		    style->bg[state_type],
		    1,
		    x,y,w,h);
  */
  cge_draw_filled_box_from_rc(xrs,
			      style,
			      state_type,
			      "background",
			      x_off,
			      y_off,
			      x,
			      y,
			      w,
			      h);

  gdk_window_set_back_pixmap(window,pm,FALSE);  

 

  //printf("set_background : %i %i %i %i %i %i\n",x_off,y_off,x,y,w,h);
 cairo_destroy(xrs);
}


GType cge_type_style = 0;

void
cge_style_register_type (GTypeModule * module)
{
  static const GTypeInfo object_info = {
    sizeof (CgeStyleClass),
    (GBaseInitFunc) NULL,
    (GBaseFinalizeFunc) NULL,
    (GClassInitFunc) cge_style_class_init,
    NULL,		/* class_finalize */
    NULL,		/* class_data */
    sizeof (CgeStyle),
    0,		/* n_preallocs */
    (GInstanceInitFunc) cge_style_init,
  };
  
  cge_type_style = g_type_module_register_type (module,
						   GTK_TYPE_STYLE,
						   "CgeStyle",
						   &object_info, 0);
}

static void
cge_style_init (CgeStyle * style)
{

}


static void
cge_style_class_init (CgeStyleClass * klass)
{
  GtkStyleClass *style_class = GTK_STYLE_CLASS (klass);
  
  parent_class = g_type_class_peek_parent (klass);


  style_class->draw_hline = draw_hline;
  style_class->draw_vline = draw_vline;  
  style_class->draw_shadow = draw_shadow; 
  style_class->draw_arrow = draw_arrow;
  style_class->draw_tab = draw_tab;
  style_class->draw_focus = draw_focus;  
  style_class->draw_box = draw_box;
  style_class->draw_flat_box = draw_flat_box;    
  style_class->draw_box_gap = draw_box_gap;
  style_class->draw_extension = draw_extension;
  style_class->draw_slider = draw_slider;  
  style_class->draw_handle = draw_handle;  
  style_class->draw_resize_grip = draw_resize_grip;  
  style_class->draw_shadow_gap = draw_shadow_gap;  
  
  style_class->draw_check = draw_check;
  style_class->draw_option = draw_option;  

  //style_class->draw_polygon = draw_polygon;
  //style_class->draw_diamond = draw_diamond;  
  //style_class->draw_layout = draw_layout;   
  
  //style_class->set_background=set_background;
}
